"""Quaxed :mod:`operator`."""

import operator
import sys
from collections.abc import Callable
from typing import Any

from quax import quaxify

__all__ = operator.__all__


def __dir__() -> list[str]:
    """List the operators."""
    return sorted(__all__)


# TODO: return type hint signature
def __getattr__(name: str) -> Callable[..., Any]:
    """Get the operator."""
    # Quaxify the operator
    out = quaxify(getattr(operator, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



__all__ = ["matmul", "matrix_transpose", "tensordot", "vecdot"]


from collections.abc import Sequence

from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def matmul(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.matmul(x1, x2)


@quaxify
def matrix_transpose(x: ArrayLike, /) -> Value:
    return array_api.matrix_transpose(x)


@quaxify
def tensordot(
    x1: ArrayLike,
    x2: ArrayLike,
    /,
    *,
    axes: int | tuple[Sequence[int], Sequence[int]] = 2,
) -> Value:
    return array_api.tensordot(x1, x2, axes=axes)


@quaxify
def vecdot(x1: ArrayLike, x2: ArrayLike, /, *, axis: int = -1) -> Value:
    return array_api.vecdot(x1, x2, axis=axis)



"""Utility functions."""

__all__ = ["all", "any"]

from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def all(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.all(x, axis=axis, keepdims=keepdims)


@quaxify
def any(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    keepdims: bool = False,
) -> Value:
    return array_api.any(x, axis=axis, keepdims=keepdims)



# ruff: noqa: E0603, F822
"""Quaxed :mod:`jax.numpy`."""

__all__ = [
    # modules
    "fft",
    "linalg",
    # contents
    "abs",
    "absolute",
    "acos",
    "acosh",
    "add",
    "all",
    "allclose",
    "amax",
    "amin",
    "angle",
    "any",
    "append",
    "apply_along_axis",
    "apply_over_axes",
    # "arange",  # in _creation_functions
    "arccos",
    "arccosh",
    "arcsin",
    "arcsinh",
    "arctan",
    "arctan2",
    "arctanh",
    "argmax",
    "argmin",
    "argpartition",
    "argsort",
    "argwhere",
    "around",
    "array",
    "array_equal",
    "array_equiv",
    "array_repr",
    "array_split",
    # "array_str",  # TODO:  why is this erroring?
    # "asarray",  # in _creation_functions
    "asin",
    "asinh",
    "astype",
    "atan",
    "atan2",
    "atanh",
    "atleast_1d",
    "atleast_2d",
    "atleast_3d",
    "average",
    "bartlett",
    "bfloat16",
    "bincount",
    "bitwise_and",
    "bitwise_count",
    "bitwise_invert",
    "bitwise_left_shift",
    "bitwise_not",
    "bitwise_or",
    "bitwise_right_shift",
    "bitwise_xor",
    "blackman",
    "block",
    "bool",
    "bool_",
    "broadcast_arrays",
    "broadcast_shapes",
    "broadcast_to",
    "c_",  # not subscriptable
    "can_cast",
    "cbrt",
    "cdouble",
    "ceil",
    "character",
    "choose",
    "clip",
    "column_stack",
    "complex128",
    "complex64",
    "complex_",
    "complexfloating",
    "compress",
    "concat",
    "concatenate",
    "conj",
    "conjugate",
    "convolve",
    "copy",
    "copysign",
    "corrcoef",
    "correlate",
    "cos",
    "cosh",
    "count_nonzero",
    "cov",
    "cross",
    "csingle",
    "cumprod",
    "cumsum",
    "deg2rad",
    "degrees",
    "delete",
    "diag",
    "diag_indices",
    "diag_indices_from",
    "diagflat",
    "diagonal",
    "diff",
    "digitize",
    "divide",
    "divmod",
    "dot",
    "double",
    "dsplit",
    "dstack",
    "dtype",
    "e",
    "ediff1d",
    "einsum",
    "einsum_path",
    "empty",
    # "empty_like",  # in _creation_functions
    "equal",
    "euler_gamma",
    "exp",
    "exp2",
    "expand_dims",
    "expm1",
    "extract",
    "eye",
    "fabs",
    "fill_diagonal",
    "finfo",
    "fix",
    "flatnonzero",
    "flexible",
    "flip",
    "fliplr",
    "flipud",
    "float16",
    "float32",
    "float64",
    "float8_e4m3b11fnuz",
    "float8_e4m3fn",
    "float8_e4m3fnuz",
    "float8_e5m2",
    "float8_e5m2fnuz",
    "float_",
    "float_power",
    "floating",
    "floor",
    "floor_divide",
    "fmax",
    "fmin",
    "fmod",
    "frexp",
    "from_dlpack",
    "frombuffer",
    # "fromfile",
    "fromfunction",
    "fromiter",
    "frompyfunc",
    "fromstring",
    # "full",  # in _creation_functions
    # "full_like",  # in _creation_functions
    "gcd",
    "generic",
    "geomspace",
    "get_printoptions",
    "gradient",
    "greater",
    "greater_equal",
    "hamming",
    "hanning",
    "heaviside",
    "histogram",
    "histogram2d",
    "histogram_bin_edges",
    "histogramdd",
    "hsplit",
    "hstack",
    "hypot",
    "i0",
    "identity",
    "iinfo",
    "imag",
    "index_exp",
    "indices",
    "inexact",
    "inf",
    "inner",
    "insert",
    "int16",
    "int32",
    "int4",
    "int64",
    "int8",
    "int_",
    "integer",
    "interp",
    "intersect1d",
    "invert",
    "isclose",
    "iscomplex",
    "iscomplexobj",
    "isdtype",
    "isfinite",
    "isin",
    "isinf",
    "isnan",
    "isneginf",
    "isposinf",
    "isreal",
    "isrealobj",
    "isscalar",
    "issubdtype",
    "iterable",
    "ix_",
    "kaiser",
    "kron",
    "lcm",
    "ldexp",
    "left_shift",
    "less",
    "less_equal",
    "lexsort",
    # "linspace",  # in _creation_functions
    "load",
    "log",
    "log10",
    "log1p",
    "log2",
    "logaddexp",
    "logaddexp2",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "logspace",
    "mask_indices",
    "matmul",
    "matrix_transpose",
    "max",
    "maximum",
    "mean",
    "median",
    # "meshgrid",  # in _creation_functions
    "mgrid",
    "min",
    "minimum",
    "mod",
    "modf",
    "moveaxis",
    "multiply",
    "nan",
    "nan_to_num",
    "nanargmax",
    "nanargmin",
    "nancumprod",
    "nancumsum",
    "nanmax",
    "nanmean",
    "nanmedian",
    "nanmin",
    "nanpercentile",
    "nanprod",
    "nanquantile",
    "nanstd",
    "nansum",
    "nanvar",
    "ndarray",
    "ndim",
    "negative",
    "newaxis",
    "nextafter",
    "nonzero",
    "not_equal",
    "number",
    "object_",
    "ogrid",
    "ones",
    # "ones_like",  # in _creation_functions
    "outer",
    "packbits",
    "pad",
    "partition",
    "percentile",
    "permute_dims",
    "pi",
    "piecewise",
    "place",
    "poly",
    "polyadd",
    "polyder",
    "polydiv",
    "polyfit",
    "polyint",
    "polymul",
    "polysub",
    "polyval",
    "positive",
    "pow",
    "power",
    "printoptions",
    "prod",
    "promote_types",
    "ptp",
    "put",
    "quantile",
    "r_",  # not subscriptable
    "rad2deg",
    "radians",
    "ravel",
    "ravel_multi_index",
    "real",
    "reciprocal",
    "remainder",
    "repeat",
    "reshape",
    "resize",
    "result_type",
    "right_shift",
    "rint",
    "roll",
    "rollaxis",
    "roots",
    "rot90",
    "round",
    "round_",
    "s_",  # not subscriptable
    "save",
    "savez",
    "searchsorted",
    "select",
    "set_printoptions",
    "setdiff1d",
    "setxor1d",
    "shape",
    "sign",
    "signbit",
    "signedinteger",
    "sin",
    "sinc",
    "single",
    "sinh",
    "size",
    "sort",
    "sort_complex",
    "split",
    "sqrt",
    "square",
    "squeeze",
    "stack",
    "std",
    "subtract",
    "sum",
    "swapaxes",
    "take",
    "take_along_axis",
    "tan",
    "tanh",
    "tensordot",
    "tile",
    "trace",
    "transpose",
    "tri",
    # "tril",  # in _creation_functions
    "tril_indices",
    "tril_indices_from",
    "trim_zeros",
    # "triu",  # in _creation_functions
    "triu_indices",
    "triu_indices_from",
    "true_divide",
    "trunc",
    # "ufunc",  # higher-order function
    "uint",
    "uint16",
    "uint32",
    "uint4",
    "uint64",
    "uint8",
    "union1d",
    "unique",
    "unique_all",
    "unique_counts",
    "unique_inverse",
    "unique_values",
    "unpackbits",
    "unravel_index",
    "unsignedinteger",
    "unwrap",
    "vander",
    "var",
    "vdot",
    "vecdot",
    # "vectorize",
    "vsplit",
    "vstack",
    "where",
    "zeros",
    # "zeros_like",  # in _creation_functions
]

import sys
from collections.abc import Callable
from typing import Any, Literal

import jax.numpy as jnp
from jaxtyping import ArrayLike

from quaxed._types import DType
from quaxed._utils import quaxify

from . import fft, linalg

# =============================================================================
# Explicit constructions
# TODO: not need to do these by specifying a `.pyi` file.
#       but right now, `_higher_order.py` needs this for mypy to pass.


@quaxify
def asarray(
    a: ArrayLike,
    dtype: DType | None = None,
    order: Literal["C", "F", "A", "K"] | None = None,
) -> ArrayLike:
    return jnp.asarray(a, dtype=dtype, order=order)


@quaxify
def expand_dims(a: ArrayLike, axis: int | tuple[int, ...]) -> ArrayLike:
    return jnp.expand_dims(a, axis=axis)


@quaxify
def squeeze(a: ArrayLike, axis: int | tuple[int, ...] | None = None) -> ArrayLike:
    return jnp.squeeze(a, axis=axis)


# =============================================================================


def __dir__() -> list[str]:
    """Return the list of names in the module."""
    return sorted(__all__)


# =============================================================================
# Automated lazy construction

# Direct transfers
_DIRECT_TRANSFER: frozenset[str] = frozenset(
    (
        "bfloat16",
        "character",
        "e",
        "euler_gamma",
        "flexible",
        "floating",
        "generic",
        "index_exp",
        "indices",
        "inexact",
        "inf",
        "int16",
        "int32",
        "int4",
        "int64",
        "int8",
        "int_",
        "integer",
        "mask_indices",
        "mgrid",
        "nan",
        "ndarray",
        "newaxis",
        "number",
        "object_",
        "ogrid",
        "pi",
        "printoptions",
        "promote_types",
        "r_",
        "s_",
        "set_printoptions",
        "signedinteger",
        "single",
        "tri",
        "tril_indices",
        "triu_indices",
        "uint",
        "uint16",
        "uint32",
        "uint4",
        "uint64",
        "uint8",
        "unsignedinteger",
    )
)


_FILTER_SPEC: dict[str, tuple[bool, ...]] = {
    "kaiser": (False, True, True),
    "swapaxes": (True, False, False),
}


def __getattr__(name: str) -> Callable[..., Any]:  # TODO: better type hint
    """Get the object from the `jax.numpy` module."""
    if name not in __all__:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    # Get the object
    jnp_obj = getattr(jnp, name)

    # Quaxify?
    out = (
        jnp_obj
        if name in _DIRECT_TRANSFER
        else quaxify(jnp_obj, filter_spec=_FILTER_SPEC.get(name, True))
    )

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



__all__ = [
    "broadcast_arrays",
    "broadcast_to",
    "concat",
    "expand_dims",
    "flip",
    "permute_dims",
    "reshape",
    "roll",
    "squeeze",
    "stack",
]

from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def broadcast_arrays(*arrays: ArrayLike) -> list[Value]:
    return array_api.broadcast_arrays(*arrays)


@quaxify
def broadcast_to(x: ArrayLike, /, shape: tuple[int, ...]) -> Value:
    return array_api.broadcast_to(x, shape)


@quaxify
def concat(
    arrays: tuple[Value, ...] | list[Value],
    /,
    *,
    axis: int | None = 0,
) -> Value:
    return array_api.concat(arrays, axis=axis)


@quaxify
def expand_dims(x: ArrayLike, /, *, axis: int = 0) -> Value:
    return array_api.expand_dims(x, axis=axis)


@quaxify
def flip(x: ArrayLike, /, *, axis: int | tuple[int, ...] | None = None) -> Value:
    return array_api.flip(x, axis=axis)


@quaxify
def permute_dims(x: ArrayLike, /, axes: tuple[int, ...]) -> Value:
    return array_api.permute_dims(x, axes=axes)


@quaxify
def reshape(
    x: ArrayLike,
    /,
    shape: tuple[int, ...],
    *,
    copy: bool | None = None,
) -> Value:
    return array_api.reshape(x, shape, copy=copy)


@quaxify
def roll(
    x: ArrayLike,
    /,
    shift: int | tuple[int],
    *,
    axis: int | tuple[int, ...] | None = None,
) -> Value:
    return array_api.roll(x, shift=shift, axis=axis)


@quaxify
def squeeze(x: ArrayLike, /, axis: int | tuple[int, ...]) -> Value:
    return array_api.squeeze(x, axis=axis)


@quaxify
def stack(arrays: tuple[Value, ...] | list[Value], /, *, axis: int = 0) -> Value:
    return array_api.stack(arrays, axis=axis)



"""Linear algebra functions."""

__all__ = [
    "cholesky",
    "cross",
    "det",
    "diagonal",
    "eigh",
    "eigvalsh",
    "inv",
    "matmul",
    "matrix_norm",
    "matrix_power",
    "matrix_rank",
    "matrix_transpose",
    "outer",
    "pinv",
    "qr",
    "slogdet",
    "solve",
    "svd",
    "svdvals",
    "tensordot",
    "trace",
    "vecdot",
    "vector_norm",
]


from collections.abc import Sequence
from typing import Literal

import jax.numpy as jnp
from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._types import DType
from quaxed._utils import quaxify


@quaxify
def cholesky(x: ArrayLike, /, *, upper: bool = False) -> Value:
    return array_api.linalg.cholesky(x, upper=upper)


@quaxify
def cross(x1: ArrayLike, x2: ArrayLike, /, *, axis: int = -1) -> Value:
    return array_api.linalg.cross(x1, x2, axis=axis)


@quaxify
def det(x: ArrayLike, /) -> Value:
    return array_api.linalg.det(x)


@quaxify
def diagonal(x: ArrayLike, /, *, offset: int = 0) -> Value:
    return array_api.linalg.diagonal(x, offset=offset)


@quaxify
def eigh(x: ArrayLike, /) -> tuple[ArrayLike]:
    return array_api.linalg.eigh(x)


@quaxify
def eigvalsh(x: ArrayLike, /) -> Value:
    return array_api.linalg.eigvalsh(x)


@quaxify
def inv(x: ArrayLike, /) -> Value:
    return array_api.linalg.inv(x)


@quaxify
def matmul(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.matmul(x1, x2)


@quaxify
def matrix_norm(
    x: ArrayLike,
    /,
    *,
    keepdims: bool = False,
    ord: int | float | Literal["fro", "nuc"] | None = "fro",
) -> Value:
    return array_api.linalg.matrix_norm(x, keepdims=keepdims, ord=ord)


@quaxify
def matrix_power(x: ArrayLike, n: int, /) -> Value:
    return array_api.linalg.matrix_power(x, n)


@quaxify
def matrix_rank(x: ArrayLike, /, *, rtol: ArrayLike | None = None) -> Value:
    return array_api.linalg.matrix_rank(x, rtol=rtol)


@quaxify
def matrix_transpose(x: ArrayLike, /) -> Value:
    return array_api.linalg.matrix_transpose(x)


@quaxify
def outer(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.linalg.outer(x1, x2)


@quaxify
def pinv(x: ArrayLike, /, *, rtol: ArrayLike | None = None) -> Value:
    return array_api.linalg.pinv(x, rtol=rtol)


@quaxify
def qr(
    x: ArrayLike,
    /,
    *,
    mode: Literal["reduced", "complete"] = "reduced",
) -> tuple[ArrayLike, ArrayLike]:
    return array_api.linalg.qr(x, mode=mode)


@quaxify
def slogdet(x: ArrayLike, /) -> tuple[Value, Value]:
    return array_api.linalg.slogdet(x)


@quaxify
def solve(x1: ArrayLike, x2: ArrayLike, /) -> Value:
    return array_api.linalg.solve(x1, x2)


@quaxify
def svd(x: ArrayLike, /, *, full_matrices: bool = True) -> tuple[Value, Value, Value]:
    return array_api.linalg.svd(x, full_matrices=full_matrices)


@quaxify
def svdvals(x: ArrayLike, /) -> Value:
    return array_api.linalg.svdvals(x)


@quaxify
def tensordot(
    x1: ArrayLike,
    x2: ArrayLike,
    /,
    *,
    axes: int | tuple[Sequence[int], Sequence[int]] = 2,
) -> Value:
    return array_api.tensordot(x1, x2, axes=axes)


@quaxify
def trace(x: ArrayLike, /, *, offset: int = 0, dtype: DType | None = None) -> Value:
    return jnp.trace(x, offset=offset, dtype=dtype)


@quaxify
def vecdot(x1: ArrayLike, x2: ArrayLike, /, *, axis: int | None = None) -> Value:
    return array_api.vecdot(x1, x2, axis=axis)


@quaxify
def vector_norm(
    x: ArrayLike,
    /,
    *,
    axis: int | tuple[int, ...] | None = None,
    keepdims: bool = False,
    ord: int | float = 2,  # pylint: disable=redefined-builtin
) -> Value:
    return array_api.linalg.vector_norm(x, axis=axis, keepdims=keepdims, ord=ord)



__all__ = ["astype", "can_cast", "finfo", "iinfo", "isdtype", "result_type"]


from jax.experimental import array_api
from jaxtyping import ArrayLike

from quaxed._types import DType
from quaxed._utils import JAX_VERSION, quaxify

if JAX_VERSION < (0, 4, 31):
    from jax.experimental.array_api._data_type_functions import FInfo
else:
    from numpy import finfo as FInfo  # noqa: N812


if JAX_VERSION < (0, 4, 27):
    from jax.experimental.array_api._data_type_functions import IInfo
else:
    from jax.experimental.array_api import iinfo as IInfo  # noqa: N812


@quaxify
def astype(x: ArrayLike, dtype: DType, /, *, copy: bool = True) -> ArrayLike:
    return array_api.astype(x, dtype, copy=copy)


@quaxify
def can_cast(from_: DType | ArrayLike, to: DType, /) -> bool:
    return array_api.can_cast(from_, to)


@quaxify
def finfo(type: DType | ArrayLike, /) -> FInfo:
    return array_api.finfo(type)


@quaxify
def iinfo(type: DType | ArrayLike, /) -> IInfo:
    return array_api.iinfo(type)


isdtype = quaxify(array_api.isdtype)


@quaxify
def result_type(*arrays_and_dtypes: ArrayLike | DType) -> DType:
    return array_api.result_type(*arrays_and_dtypes)



"""Array API creation functions."""

__all__ = [
    "arange",
    "asarray",
    "empty_like",
    "full",
    "full_like",
    "linspace",
    "meshgrid",
    "ones_like",
    "tril",
    "triu",
    "zeros_like",
]


from functools import partial
from typing import Literal, TypeVar

import jax
import jax.numpy as jnp
from jaxtyping import ArrayLike
from quax import Value

from quaxed._types import DType
from quaxed._utils import quaxify

from ._dispatch import dispatcher

T = TypeVar("T")

# =============================================================================


@dispatcher
def arange(
    start: ArrayLike,
    stop: ArrayLike | None,
    step: ArrayLike | None,
    /,
    *,
    dtype: DType | None = None,
) -> ArrayLike:
    return jnp.arange(start, stop, step, dtype=dtype)


@dispatcher  # type: ignore[no-redef]
def arange(
    start: ArrayLike,
    stop: ArrayLike | None,
    /,
    *,
    step: ArrayLike | None = None,
    dtype: DType | None = None,
) -> ArrayLike:
    # dispatch on `start`, `stop`, and `step`
    return arange(start, stop, step, dtype=dtype)


@dispatcher  # type: ignore[no-redef]
def arange(
    start: ArrayLike,
    /,
    *,
    stop: ArrayLike | None = None,
    step: ArrayLike | None = None,
    dtype: DType | None = None,
) -> ArrayLike:
    # dispatch on `start`, `stop`, and `step`
    return arange(start, stop, step, dtype=dtype)


@dispatcher  # type: ignore[no-redef]
def arange(
    *,
    start: ArrayLike,
    stop: ArrayLike | None = None,
    step: ArrayLike | None = None,
    dtype: DType | None = None,
) -> ArrayLike | Value:
    # dispatch on `start`, `stop`, and `step`
    return arange(start, stop, step, dtype=dtype)


# =============================================================================


@partial(jax.jit, static_argnames=("dtype", "order"))
@quaxify
def asarray(
    obj: ArrayLike,
    /,
    *,
    dtype: DType | None = None,
    order: Literal["C", "F", "A", "K"] | None = None,
) -> Value:
    return jnp.asarray(obj, dtype=dtype, order=order)


# =============================================================================


@dispatcher  # type: ignore[misc]
def empty_like(
    prototype: ArrayLike,
    /,
    *,
    dtype: DType | None = None,
    shape: tuple[int, ...] | None = None,
) -> ArrayLike:
    return jnp.empty_like(prototype, dtype=dtype, shape=shape)


# =============================================================================


@dispatcher
def full(
    shape: tuple[int, ...] | int,
    fill_value: ArrayLike,
    *,
    dtype: DType | None = None,
) -> ArrayLike:
    return jnp.full(shape, fill_value, dtype=dtype)


@dispatcher  # type: ignore[no-redef]
def full(
    shape: tuple[int, ...] | int,
    *,
    fill_value: ArrayLike,
    dtype: DType | None = None,
) -> ArrayLike:
    return full(shape, fill_value, dtype=dtype)


# =============================================================================


@dispatcher
def full_like(
    x: ArrayLike,
    /,
    fill_value: ArrayLike,
    *,
    dtype: DType | None = None,
    shape: tuple[int, ...] | None = None,
) -> ArrayLike:
    return jnp.full_like(x, fill_value, dtype=dtype, shape=shape)


@dispatcher  # type: ignore[no-redef]
def full_like(
    x: ArrayLike,
    *,
    fill_value: ArrayLike,
    dtype: DType | None = None,
    shape: tuple[int, ...] | None = None,
) -> ArrayLike:
    # dispatch on both `x` and `fill_value`
    return full_like.invoke(type(x), type(fill_value))(
        x, fill_value, dtype=dtype, shape=shape
    )


# =============================================================================


@dispatcher
def linspace(  # noqa: PLR0913
    start: ArrayLike,
    stop: ArrayLike,
    num: int,
    /,
    *,
    endpoint: bool = True,
    retstep: bool = False,
    dtype: DType | None = None,
    axis: int = 0,
) -> jax.Array | jax.core.Tracer | Value:
    return jnp.linspace(
        start, stop, num, endpoint=endpoint, retstep=retstep, dtype=dtype, axis=axis
    )


@dispatcher  # type: ignore[no-redef]
def linspace(  # noqa: PLR0913
    start: ArrayLike,
    stop: ArrayLike,
    /,
    *,
    num: int,
    endpoint: bool = True,
    retstep: bool = False,
    dtype: DType | None = None,
    axis: int = 0,
) -> jax.Array | jax.core.Tracer | Value:
    # dispatch on `start`, `stop`, and `num`
    return linspace(
        start, stop, num, endpoint=endpoint, retstep=retstep, dtype=dtype, axis=axis
    )


# =============================================================================


@quaxify
def meshgrid(
    *arrays: ArrayLike, copy: bool = True, sparse: bool = False, indexing: str = "xy"
) -> list[ArrayLike]:
    return jnp.meshgrid(*arrays, copy=copy, sparse=sparse, indexing=indexing)


# =============================================================================


@dispatcher  # type: ignore[misc]
def ones_like(
    x: ArrayLike, /, *, dtype: DType | None = None, shape: tuple[int, ...] | None = None
) -> ArrayLike:
    return jnp.ones_like(x, dtype=dtype, shape=shape)


# =============================================================================


@quaxify
def tril(x: ArrayLike, /, *, k: int = 0) -> ArrayLike:
    return jnp.tril(x, k=k)


# =============================================================================


# @partial(jax.jit, static_argnames=("k",))
@quaxify
def triu(x: ArrayLike, /, *, k: int = 0) -> ArrayLike:
    return jnp.triu(x, k=k)


# =============================================================================


# @partial(jax.jit, static_argnames=("dtype", "device"))
# @quaxify
@dispatcher  # type: ignore[misc]
def zeros_like(
    x: ArrayLike,
    /,
    *,
    dtype: DType | None = None,
) -> ArrayLike | jax.Array:
    return jnp.zeros_like(x, dtype=dtype)



"""Quaxed :mod:`jax.numpy`."""

# pylint: disable=redefined-builtin

from jaxtyping import install_import_hook

with install_import_hook("quaxed.numpy", None):
    from . import _core, _creation_functions, _dispatch, _higher_order
    from ._core import *  # TODO: make this lazy
    from ._creation_functions import *
    from ._dispatch import *
    from ._higher_order import *

__all__: list[str] = []
__all__ += _core.__all__
__all__ += _higher_order.__all__
__all__ += _creation_functions.__all__
__all__ += _dispatch.__all__



# ruff:noqa: F822

"""Quaxed :mod:`jax.scipy.special`."""

__all__ = [
    "bernoulli",
    "betainc",
    "betaln",
    "beta",
    "bessel_jn",
    "digamma",
    "entr",
    "erf",
    "erfc",
    "erfinv",
    "exp1",
    "expi",
    "expit",
    "expn",
    "factorial",
    "gammainc",
    "gammaincc",
    "gammaln",
    "gamma",
    "i0",
    "i0e",
    "i1",
    "i1e",
    "logit",
    "logsumexp",
    "lpmn",
    "lpmn_values",
    "multigammaln",
    "log_ndtr",
    "ndtr",
    "ndtri",
    "polygamma",
    "spence",
    "sph_harm",
    "xlogy",
    "xlog1py",
    "zeta",
    "kl_div",
    "rel_entr",
    "poch",
    "hyp1f1",
]

import sys
from collections.abc import Callable
from typing import Any

from jax.scipy import special as jsp
from quax import quaxify


def __dir__() -> list[str]:
    return sorted(__all__)


# TODO: better return type annotation
def __getattr__(name: str) -> Callable[..., Any]:
    # Quaxify the func
    func = quaxify(getattr(jsp, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, func)

    return func



"""Quaxed :mod:`jax.lax`."""

__all__ = [
    "clamp",
    "complex",
    "cumlogsumexp",
    "cummax",
    "cummin",
    "cumprod",
    "cumsum",
    "digamma",
    "erf",
    "erfc",
    "erf_inv",
    "eq",
    "select",
]


from jax import lax
from quax import quaxify

clamp = quaxify(lax.clamp)
complex = quaxify(lax.complex)
cumlogsumexp = quaxify(lax.cumlogsumexp)
cummax = quaxify(lax.cummax)
cummin = quaxify(lax.cummin)
cumprod = quaxify(lax.cumprod)
cumsum = quaxify(lax.cumsum)
digamma = quaxify(lax.digamma)
erf = quaxify(lax.erf)
erfc = quaxify(lax.erfc)
erf_inv = quaxify(lax.erf_inv)
eq = quaxify(lax.eq)
select = quaxify(lax.select)



__all__ = ["take"]

from jax.experimental import array_api
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def take(x: ArrayLike, indices: ArrayLike, /, *, axis: int | None = None) -> Value:
    return array_api.take(x, indices, axis=axis)



"""Copyright (c) 2023 Nathaniel Starkman. All rights reserved.

quaxed: Array-API JAX compatibility
"""

# pylint: disable=redefined-builtin

__all__ = ["__version__", "array_api"]

import sys
from typing import Any

import plum
from jaxtyping import ArrayLike

from . import _jax, array_api
from ._jax import *
from ._version import version as __version__

__all__ += _jax.__all__

# Simplify the display of ArrayLike
plum.activate_union_aliases()
plum.set_union_alias(ArrayLike, "ArrayLike")


def __getattr__(name: str) -> Any:  # TODO: fuller annotation
    """Forward all other attribute accesses to Quaxified JAX."""
    import jax  # pylint: disable=C0415,W0621
    from quax import quaxify  # pylint: disable=C0415,W0621

    # TODO: detect if the attribute is a function or a module.
    # If it is a function, quaxify it. If it is a module, return a proxy object
    # that quaxifies all of its attributes.
    out = quaxify(getattr(jax, name))

    # Cache the function in this module
    setattr(sys.modules[__name__], name, out)

    return out



"""FFT functions."""

__all__ = [
    "fft",
    "ifft",
    "fftn",
    "ifftn",
    "rfft",
    "irfft",
    "rfftn",
    "irfftn",
    "hfft",
    "ihfft",
    "fftfreq",
    "rfftfreq",
    "fftshift",
    "ifftshift",
]

from collections.abc import Sequence
from typing import Literal

from jax import Device
from jax.experimental.array_api import fft as _jax_fft
from jaxtyping import ArrayLike
from quax import Value

from quaxed._utils import quaxify


@quaxify
def fft(
    x: ArrayLike,
    /,
    *,
    n: int | None = None,
    axis: int = -1,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.fft(x, n=n, axis=axis, norm=norm)


@quaxify
def ifft(
    x: ArrayLike,
    /,
    *,
    n: int | None = None,
    axis: int = -1,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.ifft(x, n=n, axis=axis, norm=norm)


@quaxify
def fftn(
    x: ArrayLike,
    /,
    *,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.fftn(x, s=s, axes=axes, norm=norm)


@quaxify
def ifftn(
    x: ArrayLike,
    /,
    *,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.ifftn(x, s=s, axes=axes, norm=norm)


@quaxify
def rfft(
    x: ArrayLike,
    /,
    *,
    n: int | None = None,
    axis: int = -1,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.rfft(x, n=n, axis=axis, norm=norm)


@quaxify
def irfft(
    x: ArrayLike,
    /,
    *,
    n: int | None = None,
    axis: int = -1,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.irfft(x, n=n, axis=axis, norm=norm)


@quaxify
def rfftn(
    x: ArrayLike,
    /,
    *,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.rfftn(x, s=s, axes=axes, norm=norm)


@quaxify
def irfftn(
    x: ArrayLike,
    /,
    *,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.irfftn(x, s=s, axes=axes, norm=norm)


@quaxify
def hfft(
    x: ArrayLike,
    /,
    *,
    n: int | None = None,
    axis: int = -1,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.hfft(x, n=n, axis=axis, norm=norm)


@quaxify
def ihfft(
    x: ArrayLike,
    /,
    *,
    n: int | None = None,
    axis: int = -1,
    norm: Literal["backward", "ortho", "forward"] = "backward",
) -> Value:
    return _jax_fft.ihfft(x, n=n, axis=axis, norm=norm)


@quaxify
def fftfreq(n: int, /, *, d: float = 1.0, device: Device | None = None) -> Value:
    return _jax_fft.fftfreq(n, d=d, device=device)


@quaxify
def rfftfreq(n: int, /, *, d: float = 1.0, device: Device | None = None) -> Value:
    return _jax_fft.rfftfreq(n, d=d, device=device)


@quaxify
def fftshift(x: ArrayLike, /, *, axes: int | Sequence[int] | None = None) -> Value:
    return _jax_fft.fftshift(x, axes=axes)


@quaxify
def ifftshift(x: ArrayLike, /, *, axes: int | Sequence[int] | None = None) -> Value:
    return _jax_fft.ifftshift(x, axes=axes)



"""Quaxed :mod:`jax`."""

__all__ = [
    "device_put",
    "grad",
    "jacfwd",
]

from collections.abc import Callable, Hashable, Sequence
from typing import Any, TypeAlias

import jax
from quax import quaxify

AxisName: TypeAlias = Hashable


# =============================================================================

device_put = quaxify(jax.device_put)


def grad(  # noqa: PLR0913
    fun: Callable[..., Any],
    argnums: int | Sequence[int] = 0,
    *,
    has_aux: bool = False,
    holomorphic: bool = False,
    allow_int: bool = False,
    reduce_axes: Sequence[AxisName] = (),
    filter_spec: Any = True,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.grad`."""
    return quaxify(
        jax.grad(
            fun,
            argnums,
            has_aux=has_aux,
            holomorphic=holomorphic,
            allow_int=allow_int,
            reduce_axes=reduce_axes,
        ),
        filter_spec=filter_spec,
    )


def hessian(
    fun: Callable[..., Any],
    argnums: int | Sequence[int] = 0,
    *,
    has_aux: bool = False,
    holomorphic: bool = False,
    filter_spec: Any = True,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.hessian`."""
    return quaxify(
        jax.hessian(fun, argnums, holomorphic=holomorphic, has_aux=has_aux),
        filter_spec=filter_spec,
    )


def jacfwd(
    fun: Callable[..., Any],
    argnums: int | Sequence[int] = 0,
    *,
    has_aux: bool = False,
    holomorphic: bool = False,
    filter_spec: Any = True,
) -> Callable[..., Any]:
    """Quaxed version of :func:`jax.jacfwd`."""
    return quaxify(
        jax.jacfwd(fun, argnums, holomorphic=holomorphic, has_aux=has_aux),
        filter_spec=filter_spec,
    )



"""Copyright (c) 2023 Nathaniel Starkman. All rights reserved.

quaxed: Array-API JAX compatibility
"""

__all__: list[str] = []

from typing import Any, Protocol, TypeVar, runtime_checkable

import jax.numpy as jnp


@runtime_checkable
class DType(Protocol):
    """The dtype of an array."""

    dtype: jnp.dtype[Any]


@runtime_checkable  # TODO: need actual implementation
class SupportsBufferProtocol(Protocol):
    """Supports the buffer protocol."""


_T_co = TypeVar("_T_co", covariant=True)


@runtime_checkable
class NestedSequence(Protocol[_T_co]):
    """A nested sequence."""

    def __getitem__(self, key: int, /) -> "_T_co | NestedSequence[_T_co]": ...

    def __len__(self, /) -> int: ...
